"""External connection objects."""
import ftplib
import gzip
import os
import typing
from io import BytesIO

from riweather import utils


class NOAAFTPConnectionException(Exception):
    """Exception for bad FTP connections."""

    pass


class NOAAFTPConnection:
    """Connector to NOAA's FTP data server.

    Use this as a context manager. The connection will not occur until
    the `__enter__()` method is called, i.e. when the `with` block is entered.

    Attributes:
        ftp: The raw connected [ftplib.FTP][] instance, or `None` if the connection
            hasn't been or cannot be made.

    Examples:
        >>> from riweather import NOAAFTPConnection
        >>> with NOAAFTPConnection() as conn:
        ...     welcome = conn.ftp.getwelcome()  # see docs for ftplib.FTP
        ...     contents = conn.read_file_as_bytes("/pub/data/noaa/isd-history.txt")
        ...
        >>> "You are accessing a U.S. Government information system" in welcome
        True
        >>> contents.read(10)  # connection is closed outside of the with block
        b'Integrated'
    """

    _host = "ftp.ncei.noaa.gov"

    def __init__(self) -> None:
        """Initialize the FTP connection object."""
        self.ftp: ftplib.FTP | None = None

    def __enter__(self) -> "NOAAFTPConnection":
        """Connect to the FTP server.

        Returns:
            self
        """
        try:
            ftp = ftplib.FTP(host=self._host, timeout=30)
            ftp.login()
            self.ftp = ftp
        except OSError as e:
            raise NOAAFTPConnectionException(
                f"Could not connect to the host: {self._host}."
            ) from e
        return self

    def __exit__(self, *args) -> None:
        """Close the FTP connection gracefully.

        Raises:
            Any error encountered by `ftplib` is raised here.
        """
        try:
            self.ftp.quit()
        except ftplib.all_errors as e:
            print(e)
            self.ftp.close()

        self.ftp = None

    def read_file_as_bytes(
        self, filename: str | os.PathLike
    ) -> typing.IO | gzip.GzipFile:
        """Read a file off of the server and into a byte stream.

        Args:
            filename: The name/path of the file on the FTP server.

        Returns:
            The file contents. If `filename` ends with ".z" or ".gz", which is
                the case with many of NOAA's data files, then the results are
                decompressed automatically using [gzip][].
        """
        if self.ftp is None:
            raise NOAAFTPConnectionException(
                "FTP connection could not be established."
            ) from None
        try:
            stream = BytesIO()
            self.ftp.retrbinary("RETR {}".format(filename), stream.write)
            stream.seek(0)
        except ftplib.all_errors as e:
            raise NOAAFTPConnectionException(e) from e

        if utils.is_compressed(filename):
            return gzip.open(stream, "rb")
        else:
            return stream



"""Metadata database models."""
from sqlalchemy import Column, Float, ForeignKey, Integer, String
from sqlalchemy.orm import declarative_base, relationship

Base = declarative_base()


class Zcta(Base):
    """ZCTA database model."""

    __tablename__ = "zcta"

    id = Column(Integer, primary_key=True)
    zip = Column(String(5))
    latitude = Column(Float)
    longitude = Column(Float)
    county_id = Column(String)
    state = Column(String(2))

    def __repr__(self):
        """Pretty print ZCTA."""
        return "Zcta(id={0}, zip={1}, latitude={2}, longitude={3}, state={4})".format(
            self.id, self.zip, self.latitude, self.longitude, self.state
        )


class Station(Base):
    """Station database model."""

    __tablename__ = "station"

    id = Column(Integer, primary_key=True)
    usaf_id = Column(String, index=True)
    wban_ids = Column(String)
    recent_wban_id = Column(String(5))
    name = Column(String)
    icao_code = Column(String)
    latitude = Column(Float)
    longitude = Column(Float)
    elevation = Column(Float)
    state = Column(String(2))

    filecounts = relationship("FileCount")


class FileCount(Base):
    """Observation count model."""

    __tablename__ = "filecount"

    id = Column(Integer, primary_key=True)
    station_id = Column(Integer, ForeignKey("station.id"))
    wban_id = Column(String)
    year = Column(Integer)
    jan = Column(Integer)
    feb = Column(Integer)
    mar = Column(Integer)
    apr = Column(Integer)
    may = Column(Integer)
    jun = Column(Integer)
    jul = Column(Integer)
    aug = Column(Integer)
    sep = Column(Integer)
    oct = Column(Integer)
    nov = Column(Integer)
    dec = Column(Integer)
    count = Column(Integer)
    n_zero_months = Column(Integer)
    quality = Column(String)

    station = relationship("Station", back_populates="filecounts")



"""Weather station operations."""
import operator
from datetime import datetime

import numpy as np
import pandas as pd
import pyproj
import pytz
from pandas.tseries.frequencies import to_offset
from sqlalchemy import select

from riweather import MetadataSession
from riweather.connection import NOAAFTPConnection
from riweather.db import models

__all__ = (
    "zcta_to_lat_lon",
    "rank_stations",
    "select_station",
    "Station",
    "upsample",
    "rollup_starting",
    "rollup_ending",
    "rollup_midpoint",
    "rollup_instant",
)


def _parse_temp(s: bytes) -> float:
    value = float(s) / 10.0 if s.decode("utf-8") != "+9999" else float("nan")
    return value


def upsample(data: pd.Series | pd.DataFrame, period: str = "T"):
    """Upsample and interpolate time series data.

    Args:
        data: Time series data with a datetime index
        period: Period to upsample to. Defaults to `"T"`, which is minute-level

    Returns:
        Upsampled data

    Examples:
        >>> import pandas as pd
        >>> t = pd.Series(
        ...     [1, 2, 10],
        ...     index=pd.date_range("2023-01-01 00:01", "2023-01-01 01:05", freq="32T"),
        ... )
        >>> upsample(t, period="T").head()
        2023-01-01 00:01:00     1.00000
        2023-01-01 00:02:00     1.03125
        2023-01-01 00:03:00     1.06250
        2023-01-01 00:04:00     1.09375
        2023-01-01 00:05:00     1.12500
        Freq: T, Length: 65, dtype: float64
    """
    ts = (
        data.resample(period)
        .mean()
        .interpolate(method="linear", limit=60, limit_direction="both")
    )
    return ts


def rollup_starting(
    data: pd.Series | pd.DataFrame, period: str = "H", upsample_first: bool = True
):
    """Roll up data, labelled with the period start.

    Args:
        data: Time series data with a datetime index
        period: Period to resample to. Defaults to `"H"`, which is hourly.
        upsample_first: Perform minute-level upsampling prior to calculating
            the period average.

    Returns:
        The time series rolled up to the specified period.

    Examples:
        >>> import pandas as pd
        >>> t = pd.Series(
        ...     [1, 2, 10],
        ...     index=pd.date_range("2023-01-01 00:01", "2023-01-01 01:05", freq="32T"),
        ... )
        >>> rollup_starting(t)
        2023-01-01 00:00:00    3.207627
        2023-01-01 01:00:00    9.375000
        Freq: H, dtype: float64
    """
    if upsample_first:
        data = upsample(data, period="T")
    ts = data.resample(period, label="left", closed="left").mean()
    return ts


def rollup_ending(
    data: pd.Series | pd.DataFrame, period: str = "H", upsample_first: bool = True
):
    """Roll up data, labelled with the period end.

    Args:
        data: Time series data with a datetime index
        period: Period to resample to. Defaults to `"H"`, which is hourly.
        upsample_first: Perform minute-level upsampling prior to calculating
            the period average.

    Returns:
        The time series rolled up to the specified period.

    Examples:
        >>> import pandas as pd
        >>> t = pd.Series(
        ...     [1, 2, 10],
        ...     index=pd.date_range("2023-01-01 00:01", "2023-01-01 01:05", freq="32T"),
        ... )
        >>> rollup_ending(t)
        2023-01-01 01:00:00    3.3
        2023-01-01 02:00:00    9.5
        Freq: H, dtype: float64
    """
    if upsample_first:
        data = upsample(data, period="T")
    ts = data.resample(period, label="right", closed="right").mean()
    return ts


def rollup_midpoint(
    data: pd.Series | pd.DataFrame, period: str = "H", upsample_first: bool = True
):
    """Roll up data, labelled with the period midpoint.

    Args:
        data: Time series data with a datetime index
        period: Period to resample to. Defaults to `"H"`, which is hourly.
        upsample_first: Perform minute-level upsampling prior to calculating
            the period average.

    Returns:
        The time series rolled up to the specified period.

    Examples:
        >>> import pandas as pd
        >>> t = pd.Series(
        ...     [1, 2, 10],
        ...     index=pd.date_range("2023-01-01 00:01", "2023-01-01 01:05", freq="32T"),
        ... )
        >>> rollup_midpoint(t)
        2023-01-01 00:00:00    1.437500
        2023-01-01 01:00:00    5.661458
        Freq: H, dtype: float64
    """
    if upsample_first:
        data = upsample(data, period="T")
    half_period = to_offset(to_offset(period).delta / 2)
    ts = (
        data.shift(freq=half_period)
        .resample(period, label="left", closed="left")
        .mean()
    )
    return ts


def rollup_instant(
    data: pd.Series | pd.DataFrame, period: str = "H", upsample_first: bool = True
):
    """Roll up data, labelled with interpolated values.

    Args:
        data: Time series data with a datetime index
        period: Period to resample to. Defaults to `"H"`, which is hourly.
        upsample_first: Perform minute-level upsampling prior to returning
            a value.

    Returns:
        The time series aligned to the specified period.

    Examples:
        >>> import pandas as pd
        >>> t = pd.Series(
        ...     [1, 2, 10],
        ...     index=pd.date_range("2023-01-01 00:01", "2023-01-01 01:05", freq="32T"),
        ... )
        >>> rollup_instant(t)
        2023-01-01 00:00:00    1.00
        2023-01-01 01:00:00    8.75
        Freq: H, dtype: float64
    """
    if upsample_first:
        data = upsample(data, period="T")
    ts = data.resample(period, label="left", closed="left").first()
    return ts


class Station:  # noqa: D101
    def __init__(self, usaf_id: str, load_metadata_on_init: bool = True):
        """ISD Station object.

        Args:
            usaf_id: USAF identifier
            load_metadata_on_init: If `True`, station metadata will be retrieved
                from the local data store and loaded into the object as
                properties.

        Examples:
            >>> s = Station("720534")
            >>> print(s.name, s.latitude, s.longitude)
            ERIE MUNICIPAL AIRPORT 40.017 -105.05
        """
        self.usaf_id = usaf_id

        if load_metadata_on_init:
            self._station = self._load_metadata()
        else:
            self._station = {}

    def _load_metadata(self) -> dict:
        """Retrieve station metadata from the local data store.

        Returns:
            A Station SQLAlchemy object.
        """
        stmt = select(models.Station).where(models.Station.usaf_id == self.usaf_id)
        with MetadataSession() as session:
            station = session.scalars(stmt).first()
            station_info = {
                col: getattr(station, col) for col in station.__table__.columns.keys()
            }
            station_info["years"] = [f.year for f in station.filecounts]

        return station_info

    @property
    def wban_ids(self) -> list[str]:
        """List of valid WBAN (Weather Bureau Army Navy) identifiers."""
        return self._station.get("wban_ids", "").split(",")

    @property
    def recent_wban_id(self) -> str:
        """Most recent WBAN (Weather Bureau Army Navy) identifier."""
        return self._station.get("recent_wban_id")

    @property
    def name(self) -> str:
        """Station name."""
        return self._station.get("name")

    @property
    def icao_code(self) -> str:
        """ICAO airport code."""
        return self._station.get("icao_code")

    @property
    def latitude(self) -> float:
        """Station latitude."""
        return self._station.get("latitude")

    @property
    def longitude(self) -> float:
        """Station longitude."""
        return self._station.get("longitude")

    @property
    def elevation(self) -> float:
        """Elevation of the station, in meters."""
        return self._station.get("elevation")

    @property
    def state(self) -> str:
        """US state in which the station is located."""
        return self._station.get("state")

    @property
    def years(self) -> list[int]:
        """Years for which data exists for the station."""
        return self._station.get("years", [])

    def get_filenames(self, year: int = None) -> list[str]:
        """Construct the names of ISD files corresponding to this station.

        Args:
            year: Limit the filenames to the one corresponding to the given year.
                If `None`, filenames for all years are returned.

        Returns:
            List of filenames

        Examples:
            >>> s = Station("720534")
            >>> print(s.get_filenames(2022))
            ['/pub/data/noaa/2022/720534-00161-2022.gz']
        """
        stmt = select(models.FileCount).where(
            models.FileCount.station_id == self._station.get("id")
        )
        if year is not None:
            stmt = stmt.where(models.FileCount.year == year)

        filename_template = "/pub/data/noaa/{2}/{0}-{1}-{2}.gz"
        filenames = []
        with MetadataSession() as session:
            for row in session.scalars(stmt):
                filenames.append(
                    filename_template.format(self.usaf_id, row.wban_id, row.year)
                )

        return filenames

    def quality_report(self, year: int = None) -> pd.DataFrame | pd.Series:
        """Retrieve information on data quality.

        Args:
            year: Limit the report to information concerning the given year.
                If `None`, all years are included.

        Returns:
            Data quality report
        """
        stmt = select(models.FileCount).where(
            models.FileCount.station_id == self._station.get("id")
        )
        if year is not None:
            stmt = stmt.where(models.FileCount.year == year)

        with MetadataSession() as session:
            results = [
                {
                    "usaf_id": r.station.usaf_id,
                    "wban_id": r.wban_id,
                    "year": r.year,
                    "quality": r.quality,
                    "jan": r.jan,
                    "feb": r.feb,
                    "mar": r.mar,
                    "apr": r.apr,
                    "may": r.may,
                    "jun": r.jun,
                    "jul": r.jul,
                    "aug": r.aug,
                    "sep": r.sep,
                    "oct": r.oct,
                    "nov": r.nov,
                    "dec": r.dec,
                    "count": r.count,
                    "n_zero_months": r.n_zero_months,
                }
                for r in session.scalars(stmt).all()
            ]

        return pd.DataFrame(results).squeeze()

    def fetch_raw_temp_data(self, year: int = None, scale: str = "C") -> pd.DataFrame:
        """Retrieve raw weather data from the ISD.

        Args:
            year: Returned data will be limited to the year specified. If
                `None`, data for all years is returned.
            scale: Return the temperature in Celsius (`"C"`, the default) or
                Fahrenheit (`"F"`).

        Returns:
            A [DataFrame][pandas.DataFrame], indexed on the timestamp, with two columns:
                air temperature and dew point temperature.

        Examples:
            >>> s = Station("720534")
            >>> print(s.fetch_raw_temp_data(2022).head(2))
                                       tempC  dewC
            2022-01-01 00:15:00+00:00   -2.8  -4.0
            2022-01-01 00:35:00+00:00   -4.2  -5.5
        """
        data = []
        filenames = self.get_filenames(year)

        if scale not in ("C", "F"):
            raise ValueError('Scale must be "C" (Celsius) or "F" (Fahrenheit).')

        with NOAAFTPConnection() as conn:
            for filename in filenames:
                datastream = conn.read_file_as_bytes(filename)
                for line in datastream.readlines():
                    tempC = _parse_temp(line[87:92])
                    dewC = _parse_temp(line[93:98])
                    date_str = line[15:27].decode("utf-8")
                    dt = pytz.UTC.localize(datetime.strptime(date_str, "%Y%m%d%H%M"))
                    data.append([dt, tempC, dewC])

        timestamps, temps, dews = zip(*sorted(data), strict=True)
        ts = pd.DataFrame({"tempC": temps, "dewC": dews}, index=timestamps)

        if scale == "F":
            ts["tempF"] = ts["tempC"] * 1.8 + 32
            ts["dewF"] = ts["dewC"] * 1.8 + 32
            ts = ts.drop(["tempC", "dewC"], axis="columns")

        ts = ts.groupby(ts.index).mean()
        return ts

    def fetch_temp_data(
        self,
        year: int = None,
        value: str = None,
        scale: str = "C",
        period: str = "H",
        rollup: str = "ending",
        upsample_first: bool = True,
    ) -> pd.DataFrame | pd.Series:
        """Retrieve temperature data from the ISD.

        Args:
            year: Returned data will be limited to the year specified. If
                `None`, data for all years is returned.
            value: `"temperature"` to retrieve the air temperature only,
                or `"dew_point"` to retrieve the dew point temperature only.
                `None` returns both temperatures in a [DataFrame][pandas.DataFrame].
            scale: Return the value(s) in Celsius (`"C"`, the default) or
                Fahrenheit (`"F"`).
            period: The time step at which the data will be returned. Defaults
                to `"H"`, which corresponds to hourly data. Other possible
                values are `"30T"` or `"30min"` for half-hourly data, `"15T"`/`"15min"`
                for quarter-hourly data, and so on. See the [Pandas documentation
                on frequency strings](https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#dateoffset-objects)
                for more details on possible values.
            rollup: How to align values to the `period`. Defaults to `"ending"`,
                meaning that values over the previous time period are averaged.
            upsample_first: Whether to upsample the data to the minute level prior to
                resampling. Usually results in more accurate representations of the
                true weather data.

        Returns:
            Either a [DataFrame][pandas.DataFrame] containing both air temperature
                and dew point temperature, or, if `value` was supplied, a
                [Series][pandas.Series] containing one or the other.

        Examples:
            >>> s = Station("720534")
            >>> print(s.fetch_temp_data(2022).head(2))
                                          tempC      dewC
            2022-01-01 00:00:00+00:00 -4.298889 -5.512222
            2022-01-01 01:00:00+00:00 -6.555833 -7.688333
        """  # noqa
        if value is None:
            value = "both"
        elif value not in ("temperature", "dew_point"):
            raise ValueError('Value must be "temperature" or "dew_point"')

        if rollup not in ("starting", "ending", "midpoint", "instant"):
            raise ValueError("Invalid rollup")

        raw_ts = self.fetch_raw_temp_data(year, scale=scale)
        if rollup == "starting":
            ts = rollup_starting(raw_ts, period, upsample_first=upsample_first)
        elif rollup == "ending":
            ts = rollup_ending(raw_ts, period, upsample_first=upsample_first)
        elif rollup == "midpoint":
            ts = rollup_midpoint(raw_ts, period, upsample_first=upsample_first)
        else:  # rollup == "instant"
            ts = rollup_instant(raw_ts, period, upsample_first=upsample_first)

        if value == "temperature":
            return ts.loc[:, f"temp{scale}"]
        if value == "dew_point":
            return ts.loc[:, f"dew{scale}"]
        else:
            return ts

    def __repr__(self):
        """String representation of a Station."""
        return f'Station("{self.usaf_id}")'


def _calculate_distances(lat, lon):
    stmt = select(
        models.Station.usaf_id,
        models.Station.name,
        models.Station.latitude,
        models.Station.longitude,
    )

    with MetadataSession() as session:
        ids, names, lats, lons = zip(*session.execute(stmt), strict=True)

    target_lats = np.tile(lat, len(ids))
    target_lons = np.tile(lon, len(ids))
    geod = pyproj.Geod(ellps="WGS84")
    _, _, dists = geod.inv(
        target_lons,
        target_lats,
        lons,
        lats,
        radians=False,
    )

    data = [
        {
            "usaf_id": id,
            "name": name,
            "distance": dist,
            "latitude": lat,
            "longitude": lon,
        }
        for id, name, dist, lat, lon in zip(ids, names, dists, lats, lons, strict=True)
    ]

    return [data[i] for i in np.argsort(dists)]


def zcta_to_lat_lon(zcta: str) -> (float, float):
    """Convert zip code to lat/lon.

    Args:
        zcta: Five-digit zip code

    Returns:
        The center point of the ZCTA (Zip Code Tabulation Area).
    """
    with MetadataSession() as session:
        zcta = session.scalars(
            select(models.Zcta).where(models.Zcta.zip == zcta)
        ).first()

    return zcta.latitude, zcta.longitude


def rank_stations(
    lat: float, lon: float, *, year: int = None, max_distance_m: int = None
) -> pd.DataFrame:
    """Rank stations by distance to a point.

    Args:
        lat: Site latitude
        lon: Site longitude
        year: If specified, only include stations with data for the given year(s).
        max_distance_m: If specified, only include stations within this distance
            (in meters) from the site.

    Returns:
        A [DataFrame][pandas.DataFrame] of station information.
    """
    station_info = {info["usaf_id"]: info for info in _calculate_distances(lat, lon)}

    results = (
        select(
            models.Station.usaf_id,
            models.Station.name,
            models.FileCount.year,
            models.FileCount.quality,
        )
        .join_from(
            models.Station,
            models.FileCount,
        )
        .where(models.Station.usaf_id.in_(station_info.keys()))
    )

    data = {}
    with MetadataSession() as session:
        for row in session.execute(results):
            if row.usaf_id not in data.keys():
                data[row.usaf_id] = {
                    **station_info[row.usaf_id],
                    "years": [],
                    "quality": [],
                }

            data[row.usaf_id]["years"].append(row.year)
            data[row.usaf_id]["quality"].append(row.quality)

    data = pd.DataFrame(
        sorted(data.values(), key=operator.itemgetter("distance"))
    ).set_index("usaf_id")

    if year is not None:

        def _filter_years(x):
            if isinstance(year, list):
                return all(y in x for y in year)
            else:
                return year in x

        data = data.loc[data["years"].apply(_filter_years), :]

    if max_distance_m is not None:
        data = data.loc[data["distance"] <= max_distance_m, :]

    return data


def select_station(ranked_stations: pd.DataFrame, rank: int = 0) -> Station:
    """Return a Station object out of a ranked set of stations.

    Args:
        ranked_stations: A [DataFrame][pandas.DataFrame] returned by
            [`riweather.rank_stations`].
        rank: Which station to return. Defaults to `rank=0`, which corresponds to
            the first (i.e. nearest) station.

    Returns:
        A [`Station`][riweather.Station] object.
    """
    if len(ranked_stations) <= rank:
        raise ValueError("Rank too large, not enough stations")

    ranked_stations = ranked_stations.sort_values("distance")
    station = ranked_stations.iloc[rank]
    return Station(usaf_id=station.name)



"""Management actions on the metadata database."""
import datetime
import pathlib
import zipfile
from os import PathLike
from typing import Generator

import numpy as np
import pandas as pd
import shapefile
import shapely.geometry

from riweather import MetadataSession
from riweather.connection import NOAAFTPConnection, NOAAFTPConnectionException
from riweather.db import models


def open_zipped_shapefile(src: str | PathLike[str]) -> shapefile.Reader:
    """Open a compressed shapefile.

    Opens a zip archive, looks for a shapefile and its dependencies, and
    returns a [shapefile.Reader] object.

    Args:
        src: Path to the location of the zip archive

    Returns:
        The opened shapefile

    Raises:
        ValueError: No shapefiles, or multiple shapefiles, found in the archive
    """
    with zipfile.ZipFile(pathlib.Path(src).resolve(), "r") as archive:
        shapefiles = [
            pathlib.Path(name).stem
            for name in archive.namelist()
            if name.endswith(".shp")
        ]
        if len(shapefiles) == 0:
            raise ValueError("No shapefiles found in archive")
        elif len(shapefiles) == 1:
            sf = shapefiles[0]
        else:
            raise ValueError("Multiple shapefiles found in archive")

        shp = archive.open(".".join([sf, "shp"]))
        dbf = archive.open(".".join([sf, "dbf"]))
        try:
            shx = archive.open(".".join([sf, "shx"]))
        except ValueError:
            shx = None

    return shapefile.Reader(shp=shp, dbf=dbf, shx=shx)


def iterate_zipped_shapefile(src: str | PathLike[str]) -> Generator:
    """Iterate over the shapes and records within a zipped shapefile.

    Opens a zip archive, reads the shapefile inside, and returns each shape
    and the record associated with that shape as a dictionary.

    Args:
        src: Path to the location of the zip archive

    Yields:
        shape: A [shapely][] geometry
        record: A dictionary of data records associated with `shape`

    Examples:
        >>> for shape, record in iterate_zipped_shapefile("path/to/shapefile.zip"):  # doctest: +SKIP  # noqa
        ...     print(shape)
        ...     print(record)
    POLYGON ((...))

    {...}
    """
    with open_zipped_shapefile(pathlib.Path(src).resolve()) as sf:
        for shape_rec in sf:
            shape = shapely.geometry.shape(shape_rec.shape)
            record = shape_rec.record.as_dict()
            yield shape, record


def _map_zcta_to_state(zcta_centroid, county_metadata) -> dict:
    """Helper function to map ZCTAs to states geographically.

    Find the first instance of a county containing the given ZCTA centroid.
    If no counties are found, fall back to the first instance of a county's
    convex hull containing the ZCTA centroid. This fallback catches some, but
    probably not all, situations where the centroid is not located in any county,
    e.g. a zip code containing islands whose centroid happens to be on a lake or
    in the ocean.

    Args:
        zcta_centroid: Centroid of the target ZCTA
        county_metadata: Dictionary of county metadata (dict of dicts). Each inner
            dict must include keys "name", "state_code", "polygon", and "convex_hull".

    Returns:
        The dictionary in `county_metadata` corresponding to the matching county.
            If no match was found, the dictionary will be empty.
    """
    polygon_containment = filter(
        lambda x: zcta_centroid.within(x["polygon"]), county_metadata.values()
    )
    hit = next(polygon_containment, None)

    if hit is None:
        convex_hull_containment = filter(
            lambda x: zcta_centroid.within(x["convex_hull"]),
            county_metadata.values(),
        )
        hit = next(convex_hull_containment, {})

    return hit


def assemble_zcta_metadata(src: str | PathLike[str]) -> dict:
    """Prep ZCTA data for database insertion.

    Args:
        src: Folder containing the Census data files: "cb_2020_us_county_500k.zip"
            and "cb_2020_us_zcta520_500k.zip"

    Returns:
        A dictionary of dictionaries, keyed by 5-digit ZIP code
    """
    src = pathlib.Path(src).resolve()

    county_metadata = {}
    for shape, record in iterate_zipped_shapefile(src / "cb_2020_us_county_500k.zip"):
        county_id = record["GEOID"]
        county_metadata[county_id] = {
            "county": county_id,
            "name": record["NAMELSAD"],
            "state": record["STATE_NAME"],
            "state_code": record["STUSPS"],
            "polygon": shape,
            "convex_hull": shape.convex_hull,
            "centroid": shape.centroid,
            "latitude": shape.centroid.y,
            "longitude": shape.centroid.x,
        }

    zcta_metadata = {}
    for shape, record in iterate_zipped_shapefile(src / "cb_2020_us_zcta520_500k.zip"):
        county_record = _map_zcta_to_state(shape.centroid, county_metadata)
        zcta = record["GEOID20"]
        zcta_metadata[zcta] = {
            "zcta": zcta,
            "polygon": shape,
            "centroid": shape.centroid,
            "latitude": shape.centroid.y,
            "longitude": shape.centroid.x,
            "county_id": county_record.get("county"),
            "county_name": county_record.get("name"),
            "state": county_record.get("state_code"),
        }

    return zcta_metadata


def assemble_station_metadata(src: str | PathLike[str]) -> dict:
    """Prep weather station data for database insertion.

    Args:
        src: Folder containing the NOAA metadata file "isd-history.csv"

    Returns:
        A dictionary of dictionaries, keyed by 6-digit USAF ID
    """
    src = pathlib.Path(src).resolve()
    history = pd.read_csv(
        src / "isd-history.csv", dtype=str, parse_dates=["BEGIN", "END"]
    )

    for col in ["LAT", "LON", "ELEV(M)"]:
        # strip out "+" sign preceding value and convert to float
        history[col] = (
            history[col].str.removeprefix("+").astype(float).replace(0, np.NaN)
        )

    # h2 contains the most recent row for each ID
    h2 = history.loc[history.groupby("USAF")["END"].idxmax(), :]
    # collapse all old WBAN IDs into a single list entry per row
    h2 = h2.join(
        history.groupby("USAF")["WBAN"].apply(list).rename("wban_ids"), on="USAF"
    )

    # criteria for keeping a station:
    #   - USAF ID not equal to 999999
    #   - country is US
    #   - latitude and longitude are not missing
    h3 = (
        h2.loc[
            (h2["USAF"] != "999999")
            & (h2["CTRY"] == "US")
            & h2["LAT"].notnull()
            & h2["LON"].notnull(),
            [
                "USAF",
                "wban_ids",
                "WBAN",
                "STATION NAME",
                "ICAO",
                "LAT",
                "LON",
                "ELEV(M)",
                "STATE",
            ],
        ]
        .rename(
            columns={
                "USAF": "usaf_id",
                "WBAN": "recent_wban_id",
                "STATION NAME": "name",
                "ICAO": "icao_code",
                "LAT": "latitude",
                "LON": "longitude",
                "ELEV(M)": "elevation",
                "STATE": "state",
            }
        )
        .set_index("usaf_id")
    )

    return h3.to_dict(orient="index")


def assemble_file_metadata_from_inventory(
    src: str | PathLike[str], stations: list[str]
) -> dict:
    """Prep NOAA data file information for database insertion.

    Args:
        src: Folder containing the NOAA metadata file "isd-inventory.csv"
        stations: List of station IDs
    """
    _today = pd.to_datetime(datetime.date.today())
    src = pathlib.Path(src).resolve()
    inv: pd.DataFrame = pd.read_csv(
        src / "isd-inventory.csv", dtype={"USAF": str, "WBAN": str}
    )
    # convert column names to lowercase
    inv.columns = inv.columns.str.lower()
    # only keep stations that exist in the station metadata, and years >= 2005
    inv = inv.loc[inv["usaf"].isin(stations) & (inv["year"] >= 2005), :].copy()
    # pivot wide to long
    inv_long = inv.melt(
        id_vars=["usaf", "wban", "year"],
        var_name="month",
        value_name="count",
    ).replace(
        {
            "month": {
                "jan": 1,
                "feb": 2,
                "mar": 3,
                "apr": 4,
                "may": 5,
                "jun": 6,
                "jul": 7,
                "aug": 8,
                "sep": 9,
                "oct": 10,
                "nov": 11,
                "dec": 12,
            }
        }
    )
    # create dt column for year-month
    inv_long["day"] = 1
    inv_long["month_start_dt"] = pd.to_datetime(inv_long[["year", "month", "day"]])
    del inv_long["day"]
    # only keep year-months on or before the current month
    inv_long = inv_long.loc[inv_long["month_start_dt"] <= _today, :].copy()
    # hours in month up to today, if today falls within that month;
    # otherwise hours in total month
    inv_long["hours_in_month"] = np.where(
        (inv_long["year"] == _today.year) & (inv_long["month"] == _today.month),
        (_today - inv_long["month_start_dt"]).astype("timedelta64[h]"),
        inv_long["month_start_dt"].dt.days_in_month * 24,
    )
    q = inv_long.groupby(["usaf", "wban", "year"])[["count", "hours_in_month"]].agg(
        count=("count", "sum"),
        hours_in_year=("hours_in_month", "sum"),
        n_zero_months=("count", lambda x: x.eq(0).sum()),
    )
    # define quality criteria here
    q["quality"] = pd.Categorical(
        np.where(
            (q["count"] >= 0.9 * q["hours_in_year"]) & (q["n_zero_months"] == 0),
            "high",
            np.where(
                (q["count"] >= 0.5 * q["hours_in_year"]) & (q["n_zero_months"] <= 2),
                "medium",
                "low",
            ),
        ),
        categories=["low", "medium", "high"],
        ordered=True,
    )
    inv = inv.join(
        q[["count", "n_zero_months", "quality"]],
        on=["usaf", "wban", "year"],
        how="left",
    )
    return inv.to_dict(orient="index")


def assemble_file_metadata_from_crawl() -> list[dict]:
    """Prep NOAA data file information for database insertion.

    Traverses the FTP server and records all data files encountered, so that we can
    tell if a certain station has data for a certain year just by referencing the
    metadata database rather than hitting the server over and over again. Does not
    download any actual weather data.

    Files on the FTP server are organized like:

    /pub/data/noaa/
    |- 1901/
    ...
    |- 2022/
    |- |- [USAF-ID]-[WBAN-ID]-[YEAR].gz

    This function only walks the file system for years dating back to 2006. This is
    for two reasons. First, we generally aren't concerned with historical weather data
    more than 5-10 years old in our applications, so going way back would be a waste
    of time. Second, data coverage and quality drop off steeply the farther back you
    go - obviously there weren't as many airports in 1901 as there were in 2022, so
    we'd like to maximize our chances of pulling in decent data only.

    Returns:
        A list of dictionaries, each corresponding to a file that is present on the
        NOAA FTP server.
    """
    file_metadata = []
    with NOAAFTPConnection() as conn:
        if conn.ftp is None:
            raise NOAAFTPConnectionException(
                "FTP connection could not be established."
            ) from None
        for year in range(2006, datetime.date.today().year + 1):
            print(year)
            files = conn.ftp.mlsd(f"pub/data/noaa/{year}", ["size", "type"])
            for name, facts in files:
                if facts.get("type", "") == "file":
                    name_parts = name.split(".")[0].split("-")
                    size = int(facts.get("size", 0))
                    if len(name_parts) == 3 and size > 0:
                        file_metadata.append(
                            {
                                "usaf_id": name_parts[0],
                                "wban_id": name_parts[1],
                                "year": int(name_parts[2]),
                                "size": size,
                            }
                        )

    return file_metadata


def populate(src: str | PathLike[str]) -> None:
    """Populate the metadata database.

    Args:
        src: Folder containing the Census files `"cb_2020_us_county_500k.zip"` and
            `"cb_2020_us_zcta520_500k.zip"`, and the NOAA files `"isd-history.csv"` and
            `"isd-inventory.csv"`.
    """
    zcta_metadata = assemble_zcta_metadata(src)
    station_metadata = assemble_station_metadata(src)
    file_metadata = assemble_file_metadata_from_inventory(
        src, list(station_metadata.keys())
    )

    zcta_db_objs = [
        models.Zcta(
            zip=z["zcta"],
            latitude=z["latitude"],
            longitude=z["longitude"],
            county_id=z["county_id"],
            state=z["state"],
        )
        for z in zcta_metadata.values()
    ]

    station_db_objs = {
        usaf_id: models.Station(
            usaf_id=usaf_id,
            wban_ids=",".join(s["wban_ids"]),
            recent_wban_id=s["recent_wban_id"],
            name=s["name"],
            icao_code=s["icao_code"],
            latitude=s["latitude"],
            longitude=s["longitude"],
            elevation=s["elevation"],
            state=s["state"],
        )
        for usaf_id, s in station_metadata.items()
    }

    file_db_objs = [
        models.FileCount(
            station=station_db_objs[f["usaf"]],
            wban_id=f["wban"],
            year=f["year"],
            jan=f["jan"],
            feb=f["feb"],
            mar=f["mar"],
            apr=f["apr"],
            may=f["may"],
            jun=f["jun"],
            jul=f["jul"],
            aug=f["aug"],
            sep=f["sep"],
            oct=f["oct"],
            nov=f["nov"],
            dec=f["dec"],
            count=f["count"],
            n_zero_months=f["n_zero_months"],
            quality=f["quality"],
        )
        for f in file_metadata.values()
    ]

    with MetadataSession() as session:
        session.add_all(zcta_db_objs)
        session.add_all(station_db_objs.values())
        session.add_all(file_db_objs)
        session.commit()



"""riweather.

Grab publicly available weather data.
"""
from importlib.metadata import version, PackageNotFoundError
from importlib.resources import files

try:
    __version__ = version(__name__)
except PackageNotFoundError:
    __version__ = "unknown"

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

_dbpath = files("riweather.resources").joinpath("metadata.db")
metadata_engine = create_engine(f"sqlite+pysqlite:///{_dbpath}")

from riweather.db import Base

Base.metadata.create_all(metadata_engine)
MetadataSession = sessionmaker(metadata_engine)

from riweather.connection import NOAAFTPConnection
from riweather.stations import (
    zcta_to_lat_lon,
    rank_stations,
    select_station,
    Station,
    upsample,
    rollup_starting,
    rollup_ending,
    rollup_midpoint,
    rollup_instant,
)
from riweather.viz import plot_stations



"""Command line interface for riweather."""
import functools
import pathlib
from importlib.resources import files

import click
import requests
from sqlalchemy import create_engine

from . import __version__, connection, utils


def _cli_download_metadata(filename: str, dst: pathlib.Path) -> pathlib.Path:
    """Download metadata from NCEI/NOAA."""
    file_root = filename.split("/")[-1]
    if utils.is_compressed(file_root):
        file_root = ".".join(file_root.split(".")[:-1])
    outloc = dst / file_root

    with connection.NOAAFTPConnection() as conn:
        contents = conn.read_file_as_bytes(filename)

    with open(outloc, "wb") as f:
        f.write(contents.read())

    return outloc


def _cli_download_census(
    url: str, dst: pathlib.Path, chunk_size: int = 1024
) -> pathlib.Path:
    """Download metadata from the US Census."""
    file_root = url.split("/")[-1]
    outloc = dst / file_root

    r = requests.get(url, stream=True, timeout=30)
    n_chunks = int(int(r.headers.get("Content-Length", 0)) / chunk_size) or None

    with open(outloc, "wb") as fd:
        with click.progressbar(
            r.iter_content(chunk_size=chunk_size), length=n_chunks
        ) as bar:
            for chunk in bar:
                fd.write(chunk)

    return outloc


@click.group(context_settings={"auto_envvar_prefix": "RIWEATHER"})
@click.version_option(version=__version__)
def main() -> None:
    """Riweather makes grabbing weather data easier."""
    pass


@main.command()
@click.option(
    "-d",
    "--dst",
    required=True,
    type=click.Path(file_okay=False, writable=True, path_type=pathlib.Path),
    help="Directory where the data will be stored.",
)
def download_metadata(dst: pathlib.Path) -> None:
    """Download weather station and US geography metadata.

    Pulls files from the Internet that are necessary to (re)build the riweather
    metadata database.
    """
    secho_inprog = functools.partial(click.secho, nl=False, fg="yellow")
    secho_complete = functools.partial(click.secho, fg="blue")

    dst.mkdir(parents=True, exist_ok=True)

    secho_inprog("Downloading NCEI station metadata...")
    outloc = _cli_download_metadata("/pub/data/noaa/isd-history.csv", dst)
    secho_complete(" ---> {}".format(outloc))

    secho_inprog("Downloading NCEI station data quality report...")
    outloc = _cli_download_metadata("/pub/data/noaa/isd-inventory.csv.z", dst)
    secho_complete(" ---> {}".format(outloc))

    secho_inprog("Downloading US Census ZCTAs...", nl=True)
    outloc = _cli_download_census(
        "https://www2.census.gov/geo/tiger/GENZ2020/shp/cb_2020_us_zcta520_500k.zip",
        dst,
    )
    secho_complete(" ---> {}".format(outloc))

    secho_inprog("Downloading US Census counties...", nl=True)
    outloc = _cli_download_census(
        "https://www2.census.gov/geo/tiger/GENZ2020/shp/cb_2020_us_county_500k.zip", dst
    )
    secho_complete(" ---> {}".format(outloc))


@main.command()
@click.option(
    "-s",
    "--src",
    required=True,
    type=click.Path(file_okay=False, writable=True, path_type=pathlib.Path),
    help="Directory where the data is stored.",
)
def rebuild_db(src: pathlib.Path) -> None:
    """Drop and recreate all tables in the metadata database."""
    from riweather.db import Base
    from riweather.db.actions import populate

    dbpath = files("riweather.resources").joinpath("metadata.db")
    click.secho(
        f"Creating SQLite database at {dbpath}",
        fg="yellow",
    )

    # drop and recreate tables before populating
    metadata_engine = create_engine(f"sqlite+pysqlite:///{dbpath}")
    Base.metadata.drop_all(metadata_engine)
    Base.metadata.create_all(metadata_engine)
    click.secho("Populating database...", fg="yellow")
    populate(src)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['riweather', 'riweather.db', 'riweather.resources']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1.3,<9.0.0',
 'pandas>=2.1.0,<3.0.0',
 'pyproj>=3.4.1,<4.0.0',
 'pyshp>=2.3.1,<3.0.0',
 'requests>=2.28.1,<3.0.0',
 'shapely>=2.0.0,<3.0.0',
 'sqlalchemy>=2.0.20,<3.0.0']

extras_require = \
{'plots': ['matplotlib>=3.6.2,<4.0.0', 'folium>=0.14.0,<0.15.0']}

entry_points = \
{'console_scripts': ['riweather = riweather.cli:main']}

setup_kwargs = {
    'name': 'riweather',
    'version': '0.5.2',
    'description': 'Grab publicly available weather data',
    'long_description': '# riweather\n\n[![Tests](https://github.com/ensley-nexant/riweather/workflows/Tests/badge.svg)](https://github.com/ensley-nexant/riweather/actions?workflow=Tests)\n[![Codecov](https://codecov.io/gh/ensley-nexant/riweather/branch/main/graph/badge.svg)](https://codecov.io/gh/ensley-nexant/riweather)\n[![Release](https://github.com/ensley-nexant/riweather/actions/workflows/release.yml/badge.svg)](https://github.com/ensley-nexant/riweather/actions/workflows/release.yml)\n\nGrab publicly available weather data with `riweather`. [See the full documentation](https://ensley-nexant.github.io/riweather).\n\n## Installation\n\nInstall with pip:\n\n```\npip install riweather\n```\n\nTo create interactive maps of weather station locations, install the package along with its optional dependencies:\n\n```\npip install riweather[plots]\n```\n\n## Usage\n\nGiven a latitude and longitude, get a list of weather stations sorted from nearest to farthest from that location.\n\n```python\n>>> import riweather\n\n>>> station_rank = riweather.rank_stations(39.98, -105.13, max_distance_m=20000)\n```\n\nSelect the top station (or a different station):\n\n```python\n>>> station = riweather.select_station(station_rank, rank=0)\n```\n\nView information about that station:\n\n```python\n>>> station.name, station.usaf_id\n```\n\nAnd pull weather data from that station for a certain year.\n\n```python\n>>> station.fetch_temp_data(2022)\n```\n',
    'author': 'John Ensley',
    'author_email': 'jensley@resource-innovations.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/ensley-nexant/riweather',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
